/* Counting a counter file.
**
** Copyright (c) 2001 by Michael Newman. All rights reserved.
** Copyright (c) 2000 by Michael Newman. All rights reserved.
**
*/
#define counterEdit 4
/*
** Modification History:
**  14Jan01 MJNewman 4: Add page names.
**   7Jan01 MJNewman 3: Added quiet mode.
**  23Nov00 MJNewman 2: Added control of number of prefix 0s.
**  19Nov00 MJNewman 1: Created.
*/

//---------------------------------------------------------------------------

//#include <vcl.h>
#include <stdio.h>
#include <set>
#include <string>
#include <cstring>
#include <iostream>
#pragma hdrstop

//---------------------------------------------------------------------------

/* rev of this program */
static int majorRev = 1;
static int minorRev = 3;

#define TRUE 1
#define FALSE 0
#define BOOLEAN int
#define STRING_LENGTH 255
/* calculate the number of entrys in something */
#ifndef DIM
#define DIM(_x)          ((sizeof(_x)/sizeof(_x[0])))
#endif

/* figure the max or min of numbers */
#ifndef MAX
#define MAX(_a,_b) ((_a > _b) ? _a : _b)
#endif
#ifndef MIN
#define MIN(_a,_b) ((_a < _b) ? _a : _b)
#endif

/* various characters */
#ifndef NUL
#define NUL ((char)'\0')    /* the null character */
#endif
#define BACKSPACE 0x08
#ifndef ESC
#define ESC 27
#endif
#define UPARROW 72
#define LEFTARROW 75
#define DOWNARROW 80
#define OPEN_PAREN '('
#define CLOSE_PAREN ')'
#define OPEN_SQUARE_BRACE '['
#define CLOSE_SQUARE_BRACE ']'

class pageCount {

public:
        int count;
	std::string pageName;
        pageCount(char *argName) {
                pageName = argName;
		count = 0;
        };
        pageCount(std::string & argName) {
                pageName = argName;
		count = 0;
        };
	void resetCount() {
	    count = 0;
	    return;
	};
	void dump() const {
	    std::cout << "PageCount " << pageName << " has count " << count << '\n';
	};
	void addToCount(int number) {
	    count += number;
	    return;
	};
        friend int operator==(const pageCount &x,const pageCount &y) {
                return (x.pageName == y.pageName);
        };
        friend bool operator<(const pageCount &x,const pageCount &y) {
                return (x.pageName < y.pageName);
        };
};

std::set < pageCount > pageCounts;
std::set <pageCount>::iterator pageCountsIt;

// convert input to lower case
//
// Inputs:
//      string          string to convert to lower case
// Returns:
//      pointer to a copy of string converted to all lowe case letters valid
//      untill the next call to this routine
static char *toLowerCase(char *string)
{
    static char lowerCaseString[STRING_LENGTH];
    int i;

    for (i = 0;i < DIM(lowerCaseString);i += 1) {
        lowerCaseString[i] = tolower(string[i]);
        if (lowerCaseString[i] == NUL) return lowerCaseString;
    };
    lowerCaseString[DIM(lowerCaseString)-1] = NUL;
    return lowerCaseString;
}
        
/* Decide if a string is completely blank
 *
 *Inputs:
 *	string		string to test
 * Returns:
 *	TRUE if the string is blank
 */
static BOOLEAN isBlank(char *string)
{
    int i;
    i = 0;
    if (string == NULL) return TRUE;
    while (isspace(string[i])) {
        i += 1;
    };
    if (string[i] == NUL) return TRUE;
    return FALSE;
}

/* DOCUMENTATION:USER
** The program reads counts from an existing file and rewrites the file
** with updated counts.
**
** There are two types of counts expected in the file. One is a single count on
** a line by itself. This count is the total number of times all pages in this
** file have been read. This count can occur only once in the file.
**
** The second count is a line of the form:
**
**      $N <page name> <count>
**
** Where the <page name> is a string containing no spaces and the count is a
** number.
**
** The count file is updated only if the entire file was correctly read.
**
** DOCUMENTATION_END
*/
static const char *helpMessage =
/* DOCUMENTATION:USER */
				"COUNTER updates a file containing page counts and types the page count.\n"
				"\n"
				"The previous count file is read and the page and total counts are incremented.\n"
				"The file is written only if the previous file is correctly read.\n"
				"\n"
				"The count file contains one line with a total and lines of the form:\n\n"
				"    $N <page name> <count>\n\n"
				"The page name indicated by -nname is matched to a $N line. The count\n"
				"of that line is incremented along with the total count. The page count is\n"
				"reported when -nname is supplied. Otherwise the total count is reported.\n\n"
				"Filename for -w is countsummary.html if not given.\n"
				"\n"
				"Usage is COUNTER [-d#] [-ffilename] [-h] [-nname] [-q[-]] [-u[-]]\n"
				"\n"
				"    -d#		number of digits in output result (default 0).\n"
				"    -ffilename	file to read and then rewrite with updated count.\n"
				"    -h		emit a list of commands\n"
				"    -nname		name is the case sensitive name of a page.\n"
				"    -q[-]		process and/or update files but be quiet.\n"
				"    -u[-]		update the count file (default is to update).\n"
				"    -w[-]		turns output of web page on or off (default off).\n"
				"    -wfilename       build a web page that displays the current counts.\n"
/* DOCUMENTATION_END */
			    ;


/* Read the version information from a file
**
** Inputs
**	filename	name of input file
**	totalCount	pointer to count to update
**	found		pointer to flag set to TRUE if count found
** Returns:
**	number of errors
*/
static int readOldCounts(char *filename,int *totalCount,BOOLEAN *found)
{
    char line[STRING_LENGTH+1];
    int errorCount = 0;
    int lineNumber = 0;
    int nameCount;
    FILE *in;

    *totalCount = 0;
    *found = FALSE;
    if (filename == NULL) return 1;
    in = fopen(filename,"r");
    if (in == NULL) {
	printf("%s not found.\n",filename);
	return 1;
    };
    printf("Reading Old Counts from %s\n", filename);
    while (NULL != fgets(line,STRING_LENGTH,in)) {
        char garbage[STRING_LENGTH+1];
        char keyword[STRING_LENGTH+1];
        char subName[STRING_LENGTH+1];
        int version;
	int matches;
	lineNumber += 1;
	if (isBlank(line)) continue;
	// printf("Scanning line: %s\n",line);
        matches = sscanf(line," $%s %s %d",keyword,subName,&nameCount);
        if (matches >= 1) {
            if ((0 == strcmp(toLowerCase(keyword),"v")) && (matches == 2)) {
                matches = sscanf(line," $%s %d",keyword,&version);
                if (matches != 2) {
                    printf("Bad format version line on line %d\n",lineNumber);
                    errorCount += 1;
                };
		// printf("Found version style line keyword %s number %d\n",keyword,version);
                continue;
            };
	    // printf("Found keyword %s Subname %s file count %d\n",keyword,subName,nameCount);
	    if ((0 == strcmp(toLowerCase(keyword),"n")) && (matches == 3)) {
		std::string word(subName);
		auto search = pageCounts.find(word);
		if (search == pageCounts.end()) {
                    pageCounts.insert(word);
		    // printf("Creating entry for %s\n",subName);
		    search = pageCounts.find(word);
                };
                if (search != pageCounts.end()) {
		    // printf("Existing %s with old count %d adding %d from file\n",subName,
		    // (*search).count,nameCount);
		    pageCount copy = *search;
		    // copy.dump();
		    pageCounts.erase(search);
		    copy.addToCount(nameCount);
		    pageCounts.insert(copy);
		    search = pageCounts.find(word);
		    // ((pageCount)*search).dump();
		};
                continue;
            };
           printf("Unrecognized keyword %s on line %d\n",keyword,lineNumber);
            errorCount += 1;
            continue;
        };
	matches = sscanf(line," %d %s",&nameCount,garbage);
	if (matches == 1) {
	    if (*found) {
		printf("More than one count in %s line %d\n",filename,lineNumber);
		errorCount += 1;
	    };
	    *found = TRUE;
            *totalCount += nameCount;
            continue;
	};
	if (matches > 1) {
	    printf("Extra data found in %s on line %d\n",filename,lineNumber);
	    errorCount += 1;
	};
    };
    fclose(in);
    return errorCount;
}

// Pointer to output file created by writeNewCounts and used by outputPageCount
static FILE *outputFile;

// Output one page count
static void outputPageCount(class pageCount rCount)
{
    fprintf(outputFile,
	    "$N %s %d\n",rCount.pageName.c_str(),
	    rCount.count);
    //    printf("$N %s %d\n",rCount.pageName.c_str(), rCount.count);
    return;
}

// Output one page count for a web page
static void outputWebPageCountRow(class pageCount rCount)
{
    fprintf(outputFile,
            "<tr><td>%s</td><td>%d</td></tr>\n",rCount.pageName.c_str(),
            rCount.count);
    return;
}

/* Write counter file
**
** Inputs:
**	filename	file to output to
**	count		number to write
** Returns:
**	number of errors
*/
static int writeNewCounts(char *outputFilename,int count)
{
    int errorCount = 0;

    if (outputFilename == NULL) {
	printf("? No output file name\n");
	return 1;
    };
    outputFile = fopen(outputFilename,"w");
    if (outputFile == NULL) {
	printf("? Could not open %s\n",outputFilename);
	return 1;
    };
    fprintf(outputFile,"%d\n",count);
    for(pageCountsIt = pageCounts.begin(); pageCountsIt != pageCounts.end(); ++pageCountsIt) {
	outputPageCount(*pageCountsIt);
    };
    fclose(outputFile);
    return errorCount;
}

/* Write web file
**
** Inputs:
**	filename	file to output to
** Returns:
**	number of errors
*/
static int writeWebFile(const char *outputFilename = NULL);
static int writeWebFile(const char *outputFilename)
{
    int errorCount = 0;

    if (outputFilename == NULL) {
        printf("? No web output file name\n");
        return 1;
    };
    outputFile = fopen(outputFilename,"w");
    if (outputFile == NULL) {
        printf("? Could not open web file %s\n",outputFilename);
        return 1;
    };
    fprintf(outputFile,"<HTML><BODY>\n");
    fprintf(outputFile,"<TABLE>\n");
    fprintf(outputFile,"<TR><TH colspan=2>Page Counts</TH></TR>\n");
    fprintf(outputFile,"<TR><TH>Page</TH><TH>Count</TH></TR>\n");
    for (pageCountsIt = pageCounts.begin(); pageCountsIt != pageCounts.end(); ++pageCountsIt) {
	outputWebPageCountRow(*pageCountsIt);
    };
    fprintf(outputFile,"</TABLE>\n");
    fprintf(outputFile,"</BODY></HTML>\n");
    fclose(outputFile);
    return errorCount;
}


#pragma argsused
int main(int argc, char* argv[])
{
    int errorCount;
    char *filename = NULL;
    const char *webFilename = "countsummary.html";
    BOOLEAN webOutputEnabled = FALSE;
    int totalCount;
    int nameCount;
    int digits = 0;
    BOOLEAN found;
    BOOLEAN beQuiet = FALSE;
    BOOLEAN updateFile = TRUE;
    char *pageName = NULL;

    int i;
    if (argc < 2) {
	printf("COUNTER version %d.%d-%d\n",majorRev,minorRev,counterEdit);
	printf("%s",helpMessage);
	return (EXIT_FAILURE);
    };
    errorCount = 0;
    for (i = 1;i < argc;i += 1) {
	if (argv[i][0] == '-') {
	    switch (argv[i][1]) {
		case 'd':
		    digits = atoi(&argv[i][2]);
		    break;
		case 'h':
		    printf("COUNTER version %d.%d-%d\n",majorRev,minorRev,counterEdit);
		    printf("%s",helpMessage);
		    break;
		case 'f':
		    filename = &argv[i][2];
		    break;
                case 'n':
                    pageName = &argv[i][2];
                    break;
		case 'q':
		    beQuiet = (argv[i][2] != '-');
		    break;
                case 'u':
                    updateFile = (argv[i][2] != '-');
                    break;
                case 'w':
		    if (argv[i][2] == NUL) {
			// blank field just turns it on
			webOutputEnabled = TRUE;
			break;
		    };
		    if (argv[i][2] == '-') {
			// minus turns it off
			webOutputEnabled = FALSE;
			break;
		    };
		    webOutputEnabled = TRUE;
                    webFilename = &argv[i][2];
                    break;
		default:
		    printf("? Unrecognized selection after -: %s\n",argv[i]);
		    errorCount += 1;
		    break;
	    };
	    continue;
	};
    };
    if (errorCount > 0) {
	printf("\n%d errors encountered reading arguments.\n",errorCount);
	return (EXIT_FAILURE);
    };
    errorCount += readOldCounts(filename,&totalCount,&found);
    if (pageName != NULL) {
	std::string word(pageName);
	auto search = pageCounts.find(word);
	if (search == pageCounts.end()) {
	    // Create a count if is missing count is 0 on creation
	    pageCounts.insert(word);
	    search = pageCounts.find(word);
        };
        if (search != pageCounts.end()) {
	    pageCount copy = *search;
	    pageCounts.erase(search);
	    copy.addToCount(1);
	    pageCounts.insert(copy);
	    search = pageCounts.find(word);
            nameCount = (*search).count;
        } else {
            nameCount = 1;
        };
    };
    if ((errorCount == 0) && found && updateFile) {
	errorCount += writeNewCounts(filename,(totalCount+1));
    };
    if ((webFilename != NULL) && webOutputEnabled) {
        errorCount += writeWebFile(webFilename);
    };
    if (errorCount > 0) {
	printf("\n%d errors encountered.\n",errorCount);
	return (EXIT_FAILURE);
    };
    if (pageName == NULL) nameCount = totalCount;
    if (!beQuiet) {
	fprintf(stdout,"%0*d",digits,nameCount);
    };
    return EXIT_SUCCESS;
}

